#!/usr/bin/python
"""
    Manager.py
    
    Desc: this class uses Parallel Python to compute checksums in a distibuted
      environment. This class also interacts with file path database to retrieve
      filepaths to pass to Workers for performing checksums.
      
      
"""

#standard Python library
try:
    import cPickle as pickle
except:
    import pickle                
import datetime, getopt, hashlib, os
import pprint, random, sys, smtplib
import sqlite3, socket, threading, time

# external to Python standard library
import MySQLdb
import pp # parallel python

# our code
from configParser import Parser
from LogGenerator import LogGenerator
import ManagerServer

# number of files per run increment, to prevent disk swapping
MAX_FILES = 100000
NUM_SECS_PER_DAY = 86400

def computeChecksum(filename, workerToFilepath, algorithm):
    """
        This function gets passed to each ppserver.py (the Worker)
        program. It checks for it's own IP and uses this information
        to decide what its filepath is, then opens the file, checksums
        it and returns the result
    """
    # discover my IP using dummy socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('msu.edu', 0))
    localIP = s.getsockname()[0]
    
    filepath = workerToFilepath[localIP] # assign path using localIP
    fileToCheck = filepath+filename
    try:
        fileObj = open(fileToCheck,'r')
    except:
        checktime = time.strftime("%a, %d %b %Y %H:%M", time.localtime())
        return {'hash': "open_error", 'filename': filename, 'time': checktime}
                    
    hashFcn = hashlib.sha512()
    if algorithm == "md5":
        hashFcn = hashlib.md5()
    elif algorithm == "sha1":
        hashFcn = hashlib.sha1()
    elif algorithm == "sha224":
        hashFcn = hashlib.sha224()
    elif algorithm == "sha256":
        hashFcn = hashlib.sha256()
    elif algorithm == "sha384":
        hashFcn = hashlib.sha384()

    for line in fileObj:
        hashFcn.update(line)
    
    checktime = time.strftime("%a, %d %b %Y %H:%M", time.localtime())
    result = {'hash': binascii.b2a_hex(hashFcn.digest()), 'filename': filename, 'time': checktime}
    return result

class Manager:

        def __init__(self, algorithm=None):
            """
                Desc: load config info and set up workers
            """
            self.debug = False
            for arg in sys.argv:
                if(arg == '-d'):
                    self.debug = True

            timeStamp = time.strftime("%a, %d %b %Y %H:%M", time.localtime())
         
            configParser = Parser()
            configInfo = configParser.getConfigInfo()
            configParser.printConfigInfo()
            
            # scheduling info 
            self.runJobEveryXdays = configInfo['runJobEveryXdays'] # must be > 0
            self.jobTime = configInfo['jobStartTime'] # 4 place string
            jobHr = int(self.jobTime[:2])
            self.jobStartTimeInSecs = 60*(int(self.jobTime[2:])+jobHr*60)
            self.scheduledJob = None 
            self.lock = threading.Lock() # to make fixityTest, Add, Remove, ops atomic
            
            # set up & log file generator initialize
            if algorithm==None:
                self.algorithm = configInfo['algorithm'] # defaults to sha512
            else:
                self.algorithm = algorithm
            #end
            self.logGenerator = LogGenerator(int(configInfo['logFileSize']))
            self.emailHost = configInfo['emailHost']
           
            self.workers = [] 
            self.workerToFilepath = {} # associate IP w/ filepath for each worker
            
            # check to see that workers exist...
            if len(configInfo['workerInfo']) == 0:
                print 'Error: No workers in configuration!\n'
                sys.exit(0)
                
            for worker in configInfo['workerInfo']:
                self.workers.append(worker['IP'] + ":" + str(worker['port']))
                self.workerToFilepath[worker['IP']] = worker['path']
            
            
            self.statusMsg = "" # output string for all status msgs 
            self.servers = tuple(self.workers) # serverIPs
            
            # filepath database info 
            self.dbURL = configInfo['dbURL']
            self.dbUsername  = configInfo['dbUsername'] 
            self.dbPassword = configInfo['dbPassword'] 
            self.dbName = configInfo['dbName']
            self.dbTable = configInfo['dbTable']
            self.dbFilenameAttrib = configInfo['dbFilenameAttrib']
            self.dbFileSizeAttrib = configInfo['dbFileSizeAttrib']
            
            # email these people upon completion...
            self.notifyList = configInfo['notify']
            
            # checksum results
            self.results = []
            
            # store events here
            self.events = []
            self.errorCount = 0 
            
            # get a database connection for the manager dbase
            self.dbPath= '/dbase/'
            self.managerDBconn = None 
    
            #managerdb connection
            self.getDBconnection()
           
            self.jobServer = pp.Server(0, ppservers=self.servers)
        
        
        def getDBconnection(self):
            """
                Connect to the database. Regenerate the database if it doesn't exist.
            """
            dbase = self.dbPath+'managerdb'
            try:
                self.managerDBconn = sqlite3.connect(os.getcwd()+dbase)
                cursor = self.managerDBconn.cursor()
                cursor.execute("CREATE TABLE IF NOT EXISTS checksums(pkey TEXT PRIMARY KEY, filename TEXT, \
                            checksum TEXT, algorithm TEXT, timestamp TEXT)") 
                self.managerDBconn.commit()
            except: # recreate database if it doesn't exist...
                self.initManagerDB()
                self.managerDBconn = sqlite3.connect(os.getcwd()+dbase)
            #end
            
        def initManagerDB(self):
            """
                Create a new managerdb and initialize it. 
            """
            try: # create the dbase path
                os.mkdir(os.getcwd()+self.dbPath)
            except:
                pass
            #end
            
            dbase = self.dbPath+'managerdb'
            data = open(os.getcwd()+dbase, 'w')
            data.close()
            conn = sqlite3.connect(os.getcwd()+dbase)
            
            cursor = conn.cursor()
            cursor.execute("CREATE TABLE IF NOT EXISTS checksums(pkey TEXT PRIMARY KEY, filename TEXT, \
                            checksum TEXT, algorithm TEXT, timestamp TEXT)")
            conn.commit()                                        
            cursor.close()
            
        def run(self, currFiles=[]):
            """
                Desc: execute job, parallel python handles job assignment details
            """
            jobs = []
            self.results = [] # initialize to clear prior results
            self.fileCount = len(currFiles)
            random.shuffle(currFiles)
            
            for filename in currFiles:
                
                jobs.append(
                    self.jobServer.submit(
                            computeChecksum, 
                            (filename, self.workerToFilepath, self.algorithm), (), 
                            ("os", "hashlib", "binascii", "sys", "socket", "time"), 
                            callback = self.saveResult))
            #end
            self.jobServer.wait()

    
        def saveResult(self, result):
            """
                Append results to self.results and display progress.
            """
            if result["hash"] != "open_error":
                self.results.append(result)
            else:
                event = "FAILED TO OPEN: %s    %s" % (result['time'], result['filename'])
                self.events.append(event)
                
            progress = """\r%i of %i file(s) complete with %i file(s) skipped (%d%%).""" \
                % (len(self.results), self.fileCount, len(self.events), (100*(len(self.results) + len(self.events))/self.fileCount))
            sys.stdout.write(progress)
            sys.stdout.flush()
            #end

        
        def fixityTest(self):
            """
                This function retrieves files from the filepath database
                and computes checksums using the testfiles(..) function.
                The loop is designed to fetch rows incrementally as a
                large # of row means very slow disk swapping would need to be used.
            """
            # set the next job in case schedule jobs are set
            if self.scheduledJob != None:
                nextJobTime = self.runJobEveryXdays*NUM_SECS_PER_DAY
                self.scheduledJob = threading.Timer(nextJobTime, self.fixityTest)
            #end
            
            self.lock.acquire()
            
            # Reset workers...
            self.jobServer = pp.Server(0, ppservers=self.servers)
            
            # open connection to filename database...
            conn = MySQLdb.connect(db = self.dbName, host = self.dbURL, user = self.dbUsername, \
                                    passwd = self.dbPassword)
         
            offset = 0
            runCount = 0 
            while 1:
                conn.query("SELECT %s FROM %s LIMIT %d, %d" % (self.dbFilenameAttrib, self.dbTable, offset, MAX_FILES))
                result = conn.store_result()
                results = result.fetch_row(0)
                offset += MAX_FILES
                currFiles = []
                
                if len(results) == 0:
                    break
                
                # get the filenames....
                for row in results:
                    currFiles.append(row[0])
                
                # TEST
                #currFiles = currFiles[:80]
                
                # checksum these files, then test
                self.run(currFiles)
                self.testFiles(currFiles)
                runCount += 1
            #end
            self.logStats()
            self.lock.release()
        
        def testFiles(self, currFiles):
            """
                Compare new checksums with old and generate 
                ERROR events if checksums don't match. Files not already in 
                the managerDB will be added, generating ADD events
            """
            # in case the managerDB hasn't been activated yet...
            self.getDBconnection()
            
            cursor = self.managerDBconn.cursor()
            # retrieve stored checksums by filename & algorithm for comparison...
            newResults = []
            storedChecksums = []
            for result in self.results:
                pkey = result['filename']+'_'+self.algorithm
                cursor.execute("SELECT filename, checksum FROM checksums WHERE pkey=='%s'" % pkey)
                row = cursor.fetchone()
                if row != None:
                    storedChecksums.append(row)
                else:
                    newResults.append(result)
                #end
            #end
            
            # match filenames, compare checksums, and log events...
            for entry in storedChecksums:
                for result in self.results:
                    # check the checksums and mark as present
                    if entry[0] == result['filename']: #match filename
                        if entry[1] != result['hash']: #check for non-match betw. curr checksum & dbase
                            timestamp = result['time']
                            filename = result['filename']
                            
                            event = "ERROR: %s    %s    %s\n%s\n%s" % (timestamp, filename, self.algorithm, 'stored: '+entry[1], 'current: ' + result['hash'])
                            self.errorCount += 1
                            self.events.append(event)
                        #end
                        break
                    #end
                #end
            #end
            
            # add files to managerDB that aren't there already...
            for result in newResults:
                key = result['filename']+"_"+self.algorithm
                filename = result['filename']
                algorithm = self.algorithm
                checksum = result['hash']
                timestamp = result['time']
                
                event = "ADDED: %s    %s    %s\n%s" % (timestamp, filename, algorithm,  checksum)
                self.events.append(event)
                  
                cursor.execute("INSERT INTO checksums(pkey, filename, algorithm, checksum, timestamp) \
                                   VALUES('%s','%s','%s','%s','%s')" % (key, filename, algorithm, \
                                                        checksum, timestamp))
                self.managerDBconn.commit()
            #end
            self.logEvents()
            
        def addFile(self, someFile):
            """
                Add someFile to the managerDB
                
                Generate appropriate ADDED events
                and update the checksums table appropriately
            """
            self.lock.acquire()
            self.getDBconnection()
                
            cursor = self.managerDBconn.cursor()
           
            # check to see if the file already exists...
            pkey = someFile+'_'+self.algorithm
            cursor.execute("SELECT filename FROM checksums WHERE pkey='%s'" % pkey)
            row = cursor.fetchone()
            if row == None: # file doesn't already exist, so add it
                # compute checksum
                self.run([someFile])
           
                for result in self.results:
                    if result['hash'] != 'open_error':
                        # generate ADDED log events
                        pkey = result['filename']+'_'+self.algorithm
                        filename = result['filename']
                        algorithm = self.algorithm
                        checksum = result['hash']
                        timestamp = result['time']
                        
                        event = "ADDED: %s    %s    %s\n%s" % (timestamp, filename, algorithm,  checksum)
                        self.events.append(event)
                    
                        # update the database...
                        cursor.execute("INSERT INTO checksums(pkey, filename, algorithm, checksum, timestamp) \
                                VALUES('%s','%s','%s','%s', '%s')" % (pkey, filename, algorithm, \
                                        checksum, timestamp))
                    #end
                #end
            #end
            self.managerDBconn.commit()
            self.logEvents()
            self.lock.release()
            
        def removeFile(self, someFile):
            """
                Given a string representing a filename or filepath, remove it from
                managerdb
            """
            self.lock.acquire()
            self.getDBconnection()
                
            cursor = self.managerDBconn.cursor()
            
            cursor.execute("SELECT filename, algorithm, checksum  FROM checksums WHERE filename='%s'" % someFile)
            results = cursor.fetchall()
            removedEntries = []
            if results != None: # file exists, so add to remove list
                for result in results:
                    removedEntries.append(result)
            
            for entry in removedEntries:
                # generate REMOVED log events
                timestamp = time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime())
                filename = entry[0]
                algorithm = entry[1]
                checksum = entry[2]
                
                event = "REMOVED: %s    %s    %s    %s" % (timestamp, filename, algorithm, checksum)
                self.events.append(event)
                
                # update the database...
                cursor.execute("DELETE FROM checksums WHERE filename='%s'" % filename)
            #end
            self.managerDBconn.commit()
            self.logEvents()
            self.lock.release()
            
        def logEvents(self):
            """
                Put all log events into data files, then clear 
                all current events if they've reached the MAX_FILES 
                threshold.
            """
            if len(self.events) >= MAX_FILES:
                print # print to next line after progress bar
                timestamp = time.strftime("%a_%d_%b_%Y_%H%M", time.localtime())
                self.logGenerator.saveLog(self.events, timestamp)
                self.events = []
                self.errorCount = 0
            #end
            
        def forceLogDump(self):
            """
                Output events to log files regardless of number. This 
                ensures log files get written out even if too few have
                been created.
            """
            if len(self.events) > 0:
                print # print to next line after progress bar
                timestamp = time.strftime("%a_%d_%b_%Y_%H%M", time.localtime())
                self.logGenerator.saveLog(self.events, timestamp)
                self.events = []
                self.errorCount = 0
            #end
        
        def logStats(self):
            """
                Save the statistics from the Worker in a log file
            """
            if self.errorCount==0:
                self.statusMsg += "\nNo errors were found.\n\n"
            else:
                self.statusMsg += "\n%d errors were found\n\n" % self.errorCount
            #end
            
            stats =  self.jobServer.get_stats()
            self.statusMsg += 'STATS BY MACHINE....\n'
            self.statusMsg += 'Number of CPUs      Time (in mins)    Number of Files              Worker\n'
            for key in stats.keys():
                stat = stats[key]
                if stat.rworker != None:#'Number of CPUs      Time (in mins)    Number of Files              Worker\n'
                    self.statusMsg +=    '            %d                 %.2f                %d                  %s\n' % (stat.ncpus, stat.time/60.0, stat.njobs, stat.rworker.id)
                #end
            #end
            # write LOG stats, call email, and clear current statusMsg
            self.statusMsg += "\nUsing %s as the checksum algorithm\n" % self.algorithm
            timestamp = time.strftime("%a_%d_%b_%Y_%H%M", time.localtime())
            self.logGenerator.saveLog([self.statusMsg], timestamp+'_STATS')
            self.emailNotification()
            self.statusMsg = ""
            
        
        def runSchedule(self):
            """
                Run jobs according to the schedule in config file.
                Ex. every 10 days at 1900
            """
            if self.scheduledJob == None:
                currTime = time.localtime()
                currHr = currTime[3]
                currTimeInSecs = 60*(currTime[4]+60*currHr)
                
                nextJobTime = 1

                if currTimeInSecs > self.jobStartTimeInSecs:
                    nextJobTime = (NUM_SECS_PER_DAY-currTimeInSecs)+self.jobStartTimeInSecs
                else:
                    nextJobTime = self.jobStartTimeInSecs-currTimeInSecs
                
                self.scheduledJob = threading.Timer(nextJobTime, self.fixityTest)
                self.scheduledJob.start()
            #end
      
        def emailNotification(self):
            """
                Sends the Manager's self.statusMsg field to any admins in the config using the host 
                    named in the config
            """
            sender = "KORA_EVENT_LOGGER"
            try:
                for recvr in self.notifyList:
                    headers = "From: %s\r\nTo: %s\r\nSubject: %s\r\n\r\n" % (sender, recvr, "FIXITY TEST COMPLETE")
                    message = headers + self.statusMsg
                    mailServer = smtplib.SMTP(self.emailHost)
                    mailServer.sendmail(sender, recvr, message)
                    mailServer.quit()
                #end
            except:
                print 'Email Notification Failed' 
            #end


# Runnable script
myManager = Manager()
# if cmd line args don't enter loop
if len(sys.argv) > 1:
    try:
        optlist, args = getopt.getopt(sys.argv[1:], 'a:r:fs')
    except getopt.error, msg:
        print msg
        print 'Valid options are: '
        print "   -a <filename> to add a file "
        print "   -r <filename> to remove a file "
        print "   -f            run a fixity test"
        print "   -s            run the scheduled fixity test"
        print 
        sys.exit(0)
    
    for opt in optlist:
        if opt[0] == '-a': # add
            myManager.addFile(opt[1])
        elif opt[0] == '-r': # remove
            myManager.removeFile(opt[1])
        elif opt[0] == '-f': # fixity
            myManager.fixityTest()
        elif opt[0] == '-s': # schedule
            myManager.runSchedule()
else:
    while(1):
        print "\r\n \r\n"
        print "//////////////////////////////"
        print "//    KORA EVENT LOGGER     //"
        print "//                          //"
        print "//                          //"
        print "// 0 => Run Fixity Test     //"
        print "// 1 => Run Schedule        //"
        print "// 2 => Add a File          //"
        print "// 3 => Remove a File       //"
        print "// 4 => Start XML-RPC server//"
        print "// otherwise => Quit        //"
        print "//////////////////////////////"
        result = raw_input("What would you like to do? ")

        if result == '0': # single test
            myManager.fixityTest()
            myManager.forceLogDump()
        elif result == '1': # run scheduler
            myManager.runSchedule()
            myManager.forceLogDump()
        elif result == '2': # add a file
            newFilename = raw_input("Enter a File to Add: ")
            myManager.addFile(newFilename)
            myManager.forceLogDump()
        elif result == '3': # remove a file
            removeFilename = raw_input("Enter a File to Remove: ")
            myManager.removeFile(removeFilename)
            myManager.forceLogDump()
        elif result == '4': # start XML-RPC server
            try:
                port = int(raw_input("Enter port # to listen on:"))
                ManagerServer.startServer(myManager, port)
                #myManager.forceLogDump()
            except KeyboardInterrupt:
                print 'Closing server...'
        else:
            break



